==> ben-2.08.tar.gz: ben/arg.c <==
/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2019  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "xmem.h"
#include "buffer.h"
#include "strlist.h"
#include "config.h"
#include "arg.h"


// ***************************************************************************
// 
// ***************************************************************************
#define MSG(...)	fprintf(stderr, __VA_ARGS__)

// ***************************************************************************
// 
// ***************************************************************************
void arg_init(args *a, int argc, char **argv)
{
	int t;
	char *d;
	
	a->argc = argc;
	a->argv = xmalloc(argc * sizeof(char **));
	
	buffer_init(&a->buff, 0);
	strlist_init(&a->str);

	for (t = 0; t < argc; t++) {
		d = xstrdup(argv[t]);
		arg_hook(a, d);
		a->argv[t] = d;
	}
}

void arg_clear(args *a)
{
	size_t i, n;
	char **l;
	
	l = strlist_data(&a->str);
	n = a->str.n;
	
	for (i = 0; i < n; i++) {
		xfree(l[i]);
	}
	
	xfree(a->argv);
	buffer_clear(&a->buff);
	strlist_clear(&a->str);
}

// ***************************************************************************
// 
// ***************************************************************************
void arg_hook(args *a, char *str)
{
	strlist_append_one(&a->str, str);
}

// ***************************************************************************
// 
// ***************************************************************************
static int arg_find(args *a, char *s, char *l, char **val)
{
	size_t sl, ll;
	char *v, *e;
	int t;
	
	if ((!s) && (!l))
		return(0);
	
	sl = (s) ? strlen(s) : 0;
	ll = (l) ? strlen(l) : 0;
	
	v = e = NULL;
	
	for (t = 1; t < a->argc; t++) {
		v = a->argv[t];
		
		// empty or already processed
		if (v[0] == 0)
			continue;
		
		// long form
		if ((v[0] == '-') && (v[1] == '-')) {
			if (!l)
				continue;
			if (l[0] == 0)
				continue;
			
			if (strncmp(v + 2, l, ll))
				continue;
			
			e = v + 2 + ll;
			
			if (e[0] == 0)
				break;
			
			if (e[0] == '=') {
				e++;
				break;
			}
			
			continue;
		}
		
		// short form
		if ((v[0] == '-') && (v[1] != 0)) {
			if (!s)
				continue;
			if (s[0] == 0)
				continue;
			
			if (strncmp(v + 1, s, sl))
				continue;
			
			e = v + 1 + sl;
			
			break;
		}
		
		// naked
		if (((s) && (s[0] == 0)) || ((l) && (l[0] == 0))) {
			e = v;
			break;
		}
	}
	
	// not found
	if (t >= a->argc)
		return(0);
	
	// no value
	if (!val) {
		if (e[0] != 0) {
			//MSG("%s: no value expected (%s)\n", v, e);
			return(0);
		}
		
		a->argv[t] = "";
		return(1);
	}
	
	// value came with argument
	if (e[0] != 0) {
		*val = e;
		
		a->argv[t] = "";
		return(1);
	}
	
	// value comes with next argument
	if (t + 1 >= a->argc) {
		//MSG("%s: argument expected\n", v);
		return(0);
	}
	
	e = a->argv[t + 1];
	
	if ((e[0] == 0) || ((e[0] == '-') && (e[1] != 0))) {
		//MSG("%s: argument expected\n", v);
		return(0);
	}
	
	*val = e;
	
	a->argv[t] = "";
	a->argv[t + 1] = "";
	
	return(1);
}

// ***************************************************************************
// 
// ***************************************************************************
int arg_remain(args *a)
{
	int t, cnt;
	char *v;
	
	cnt = 0;
	
	for (t = 1; t < a->argc; t++) {
		v = a->argv[t];
		
		if (v[0] != 0)
			cnt++;
	}
	
	return(cnt);
}


int arg_finish(args *a)
{
	int t;
	char *v;
	
	for (t = 1; t < a->argc; t++) {
		v = a->argv[t];
		
		if (v[0] != 0) {
			MSG("%s: Unknown argument\n", v);
			return(-1);
		}
	}
	
	return(0);
}


// ***************************************************************************
// 
// ***************************************************************************
int arg_switch(args *a, char *sd, char *ld)
{
	return(arg_find(a, sd, ld, NULL));
}

// ***************************************************************************
// 
// ***************************************************************************
char *arg_str(args *a,
	char *sd, char *ld, char *sf, char *lf,
	char *def_val)
{
	char *val;
	char *d;
	size_t len;
	
	if (arg_find(a, sd, ld, &val)) {
		return(val);
	}
	
	if (arg_find(a, sf, lf, &val)) {
		if (buffer_load_path(&a->buff, val, conf.file_size_max))
			exit(1);
		
		len = a->buff.size;
		
		d = xmalloc((len + 1) * sizeof(char));
		
		memcpy(d, a->buff.data, len * sizeof(char));
		
		d[len] = 0;
		
		arg_hook(a, d);
		
		return(d);
	}
	
	return(def_val);
}

// ***************************************************************************
// 
// ***************************************************************************
int arg_int(args *a, char *sd, char *ld, int def_val)
{
	char *val;
	char *e;
	long r;
	
	val = arg_str(a, sd, ld, NULL, NULL, NULL);
	
	if (!val)
		return(def_val);
	
	r = strtol(val, &e, 0);
	
	if ((e == val) || (*e != 0)) {
		MSG("-%s/--%s: %s: Unable to parse value\n",
			sd, ld, val);
		exit(1);
	}
	
	return(r);
}

// ***************************************************************************
// 
// ***************************************************************************
int arg_strlist(args *a,
	char *sd, char *ld, char *sf, char *lf,
	strlist *l)
{
	char *val;
	size_t n0;
	
	n0 = l->n;
	
	while (1) {
		val = arg_str(a, NULL, NULL, sf, lf, NULL);
		
		if (!val)
			break;

		strlist_parse(l, val);
	}
	
	while (1) {
		val = arg_str(a, sd, ld, NULL, NULL, NULL);
		
		if (!val)
			break;
		
		strlist_append_one(l, val);
	}
	
	return(l->n - n0);
}

==> ben-2.17.tar.gz: ben/arg.c <==
/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "xmem.h"
#include "buffer.h"
#include "strlist.h"
#include "config.h"
#include "arg.h"


// ***************************************************************************
// 
// ***************************************************************************
#define MSG(...)	fprintf(stderr, __VA_ARGS__)

// ***************************************************************************
// 
// ***************************************************************************
void arg_init(args *a, int argc, char **argv)
{
	int t;
	char *d;
	
	a->argc = argc;
	a->argv = xmalloc(argc * sizeof(char **));
	
	buffer_init(&a->buff, 0);
	strlist_init(&a->str);

	for (t = 0; t < argc; t++) {
		d = xstrdup(argv[t]);
		arg_hook(a, d);
		a->argv[t] = d;
	}
}

void arg_clear(args *a)
{
	size_t i, n;
	char **l;
	
	l = strlist_data(&a->str);
	n = a->str.n;
	
	for (i = 0; i < n; i++) {
		xfree(l[i]);
	}
	
	xfree(a->argv);
	buffer_clear(&a->buff);
	strlist_clear(&a->str);
}

// ***************************************************************************
// 
// ***************************************************************************
void arg_hook(args *a, char *str)
{
	strlist_append_one(&a->str, str);
}

// ***************************************************************************
// 
// ***************************************************************************
static int arg_find(args *a, char *s, char *l, char **val)
{
	size_t sl, ll;
	char *v, *e;
	int t;
	
	if ((!s) && (!l))
		return(0);
	
	sl = (s) ? strlen(s) : 0;
	ll = (l) ? strlen(l) : 0;
	
	v = e = NULL;
	
	for (t = 1; t < a->argc; t++) {
		v = a->argv[t];
		
		// empty or already processed
		if (v[0] == 0)
			continue;
		
		// long form
		if ((v[0] == '-') && (v[1] == '-')) {
			if (!l)
				continue;
			if (l[0] == 0)
				continue;
			
			if (strncmp(v + 2, l, ll))
				continue;
			
			e = v + 2 + ll;
			
			if (e[0] == 0)
				break;
			
			if (e[0] == '=') {
				e++;
				break;
			}
			
			continue;
		}
		
		// short form
		if ((v[0] == '-') && (v[1] != 0)) {
			if (!s)
				continue;
			if (s[0] == 0)
				continue;
			
			if (strncmp(v + 1, s, sl))
				continue;
			
			e = v + 1 + sl;
			
			break;
		}
		
		// naked
		if (((s) && (s[0] == 0)) || ((l) && (l[0] == 0))) {
			e = v;
			break;
		}
	}
	
	// not found
	if (t >= a->argc)
		return(0);
	
	// no value
	if (!val) {
		if (e[0] != 0) {
			//MSG("%s: no value expected (%s)\n", v, e);
			return(0);
		}
		
		a->argv[t] = "";
		return(1);
	}
	
	// value came with argument
	if (e[0] != 0) {
		*val = e;
		
		a->argv[t] = "";
		return(1);
	}
	
	// value comes with next argument
	if (t + 1 >= a->argc) {
		//MSG("%s: argument expected\n", v);
		return(0);
	}
	
	e = a->argv[t + 1];
	
	if ((e[0] == 0) || ((e[0] == '-') && (e[1] != 0))) {
		//MSG("%s: argument expected\n", v);
		return(0);
	}
	
	*val = e;
	
	a->argv[t] = "";
	a->argv[t + 1] = "";
	
	return(1);
}

// ***************************************************************************
// 
// ***************************************************************************
int arg_remain(args *a)
{
	int t, cnt;
	char *v;
	
	cnt = 0;
	
	for (t = 1; t < a->argc; t++) {
		v = a->argv[t];
		
		if (v[0] != 0)
			cnt++;
	}
	
	return(cnt);
}


int arg_finish(args *a)
{
	int t;
	char *v;
	
	for (t = 1; t < a->argc; t++) {
		v = a->argv[t];
		
		if (v[0] != 0) {
			MSG("%s: Unknown, duplicate or missing argument\n",
				v);
			return(-1);
		}
	}
	
	return(0);
}


// ***************************************************************************
// 
// ***************************************************************************
int arg_switch(args *a, char *sd, char *ld)
{
	return(arg_find(a, sd, ld, NULL));
}

// ***************************************************************************
// 
// ***************************************************************************
char *arg_str(args *a,
	char *sd, char *ld, char *sf, char *lf,
	char *def_val)
{
	char *val;
	char *d;
	size_t len;
	
	if (arg_find(a, sd, ld, &val)) {
		return(val);
	}
	
	if (arg_find(a, sf, lf, &val)) {
		if (buffer_load_path(&a->buff, val, conf.file_size_max))
			exit(1);
		
		len = a->buff.size;
		
		d = xmalloc((len + 1) * sizeof(char));
		
		memcpy(d, a->buff.data, len * sizeof(char));
		
		d[len] = 0;
		
		arg_hook(a, d);
		
		return(d);
	}
	
	return(def_val);
}

// ***************************************************************************
// 
// ***************************************************************************
int arg_int(args *a, char *sd, char *ld, int def_val)
{
	char *val;
	char *e;
	long r;
	
	val = arg_str(a, sd, ld, NULL, NULL, NULL);
	
	if (!val)
		return(def_val);
	
	r = strtol(val, &e, 0);
	
	if ((e == val) || (*e != 0)) {
		MSG("-%s/--%s: %s: Unable to parse value\n",
			sd, ld, val);
		exit(1);
	}
	
	return(r);
}

// ***************************************************************************
// 
// ***************************************************************************
int arg_strlist(args *a,
	char *sd, char *ld, char *sf, char *lf,
	strlist *l)
{
	char *val;
	size_t n0;
	
	n0 = l->n;
	
	while (1) {
		val = arg_str(a, NULL, NULL, sf, lf, NULL);
		
		if (!val)
			break;

		strlist_parse(l, val);
	}
	
	while (1) {
		val = arg_str(a, sd, ld, NULL, NULL, NULL);
		
		if (!val)
			break;
		
		strlist_append_one(l, val);
	}
	
	return(l->n - n0);
}
